import inspect

import anyio
import rich
import typer

from unpage.cli._app import app
from unpage.telemetry import client as telemetry


@app.command()
def version(json: bool = typer.Option(False, help="return json output")) -> None:
    """
    Display the version of the Unpage CLI.
    """

    async def _version() -> None:
        await telemetry.send_event(
            {
                "command": "version",
                "json_output": json,
            }
        )

        from dspy import __version__ as dspy_version
        from dspy.adapters.types.tool import Tool

        from unpage import __version__

        dspy_parse_function_tool_source = inspect.getsource(Tool._parse_function)
        dspy_tool_no_input_args_bugfix_present = all(
            " is not None else " in line
            for line in dspy_parse_function_tool_source.splitlines()[-3:-1]
        )

        if json:
            rich.print_json(
                data={
                    "unpage": __version__,
                    "dspy": dspy_version,
                    "dspy_tool_no_input_args_bugfix_present": dspy_tool_no_input_args_bugfix_present,
                }
            )
            return
        print(
            f"unpage {__version__} (dspy {dspy_version} {dspy_tool_no_input_args_bugfix_present=})"
        )

    anyio.run(_version)



from pydantic_yaml import parse_yaml_raw_as
import pytest
from fastmcp import Client

from unpage.agent.analysis import Agent
from unpage.agent.utils import get_agent_template, load_agent


@pytest.mark.asyncio
async def test_default_template_has_all_tools(mcp_client: Client) -> None:
    tools = await mcp_client.list_tools()
    server_tool_names = [tool.name for tool in tools]
    default_template = get_agent_template("default")
    default_template_tools = parse_yaml_raw_as(Agent, default_template).tools
    assert sorted(default_template_tools) == sorted(server_tool_names)



import os

import anyio
import typer
from rich import print

from unpage.cli.agent._app import agent_app
from unpage.cli.agent.actions import create_agent
from unpage.cli.options import PROFILE_OPTION
from unpage.telemetry import client as telemetry
from unpage.telemetry import hash_value, prepare_profile_for_telemetry
from unpage.utils import edit_file


@agent_app.command()
def create(
    agent_name: str = typer.Argument(..., help="The name of the agent to create"),
    profile: str = PROFILE_OPTION,
    overwrite: bool = typer.Option(False, help="Overwrite the agent file if it already exists"),
    template: str = typer.Option("default", help="The template to use to create the agent file"),
    editor: str = typer.Option(
        os.environ.get("EDITOR"), help="The editor to use to open the agent file"
    ),
    no_edit: bool = typer.Option(False, help="Do not open the agent file in your editor"),
) -> None:
    """Create a new agent configuration file and open it in your editor."""

    async def _create() -> None:
        await telemetry.send_event(
            {
                "command": "agent create",
                "agent_name_sha256": hash_value(agent_name),
                **prepare_profile_for_telemetry(profile),
                "overwrite": overwrite,
                "template": template,
                "editor": editor,
                "no_edit": no_edit,
            }
        )
        agent_file = create_agent(
            agent_name=agent_name,
            profile=profile,
            overwrite=overwrite,
            template=template,
        )
        # Open the file in the user's editor
        if editor and not no_edit:
            try:
                await edit_file(agent_file, editor)
            except ValueError as ex:
                print(
                    "[red]No editor specified. Set the $EDITOR environment variable or use --editor option.[/red]"
                )
                print(f"[blue]Please manually open {str(agent_file)!r} in your editor.[/blue]")
                raise typer.Exit() from ex

    anyio.run(_create)



import typer

from unpage.cli.mcp.client.logs.app import client_logs_app

client_app = typer.Typer(
    help="Debugging tools for clients of the Unpage MCP Server", no_args_is_help=True
)
client_app.add_typer(client_logs_app, name="logs")



"""
Vendored dependencies for unpage.
"""

# Re-export vendored packages
import sys
from pathlib import Path

# Add the vendor directory to sys.path
vendor_dir = Path(__file__).parent
if str(vendor_dir) not in sys.path:
    sys.path.insert(0, str(vendor_dir))



from pathlib import Path

import typer

from unpage.agent.utils import get_agent_template
from unpage.config.utils import get_config_dir


def create_agent(agent_name: str, profile: str, overwrite: bool, template: str) -> Path:
    # Create the default YAML content
    try:
        agent_template = get_agent_template(template)
    except FileNotFoundError as ex:
        print(
            f"Template '{template}' not found at {Path(__file__).parent / 'templates' / f'{template}.yaml'}"
        )
        raise typer.Abort() from ex

    # Get the config directory for the profile
    config_dir = get_config_dir(profile, create=True)

    # Create the agents directory if it doesn't exist
    agents_dir = config_dir / "agents"
    agents_dir.mkdir(exist_ok=True)

    # Create the agent file path
    agent_file = agents_dir / f"{agent_name}.yaml"

    # Check if the agent file already exists
    if agent_file.exists():
        if overwrite:
            print(f"Overwriting agent '{agent_name}' at {agent_file}")
        else:
            print(f"Agent '{agent_name}' already exists at {agent_file}")
            if not typer.confirm("Do you want to overwrite it?"):
                raise typer.Abort()

    # Write the YAML content to the file
    agent_file.write_text(agent_template, encoding="utf-8")

    print(f"Created agent configuration at {agent_file}")
    return agent_file



import os
import signal

import anyio

from unpage.cli.graph._app import graph_app
from unpage.cli.graph._background import cleanup_pid_file, get_pid_file, is_process_running
from unpage.cli.options import PROFILE_OPTION
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@graph_app.command()
def stop(profile: str = PROFILE_OPTION) -> None:
    """Stop running graph build"""

    async def _stop() -> None:
        await telemetry.send_event(
            {
                "command": "graph stop",
                **prepare_profile_for_telemetry(profile),
            }
        )
        _stop_graph()

    def _stop_graph() -> None:
        pid_file = get_pid_file(profile)

        if not pid_file.exists():
            print(f"No graph build running for profile '{profile}'")
            return

        try:
            pid = int(pid_file.read_text().strip())
            if is_process_running(pid):
                print(f"Stopping graph build for profile '{profile}' (PID: {pid})...")
                os.kill(pid, signal.SIGTERM)
                print("Graph build stopped successfully")
            else:
                print(f"Process not found for profile '{profile}', cleaning up stale PID file...")
            cleanup_pid_file(profile)
        except (ValueError, ProcessLookupError):
            cleanup_pid_file(profile)
            print(f"No running process found for profile '{profile}', cleaned up stale PID file")

    anyio.run(_stop)



import os

import anyio
import typer

from unpage.agent.app import listen, settings
from unpage.cli.agent._app import agent_app
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@agent_app.command()
def serve(
    host: str = typer.Option(settings.UNPAGE_HOST, help="The host to bind to"),
    port: int = typer.Option(settings.UNPAGE_PORT, help="The port to bind to"),
    workers: int = typer.Option(settings.UNPAGE_WORKERS, help="The number of workers to use"),
    profile: str = typer.Option(settings.UNPAGE_PROFILE, help="The profile to use"),
    reload: bool = typer.Option(
        settings.UNPAGE_RELOAD, help="Reload the server when the code changes"
    ),
    tunnel: bool = typer.Option(settings.UNPAGE_TUNNEL, help="Tunnel the server through ngrok"),
    ngrok_token: str = typer.Option(
        os.getenv("NGROK_TOKEN"),
        help="The ngrok token to use to tunnel the server",
        show_default=False,
    ),
    ngrok_domain: str = typer.Option(
        settings.NGROK_DOMAIN, help="The ngrok domain to use to tunnel the server"
    ),
) -> None:
    """Run the Unpage Agent server, which loads all agents and routes between them. This is intended to be a webhook receiver for PagerDuty."""

    async def _serve() -> None:
        await telemetry.send_event(
            {
                "command": "agent serve",
                "host": host
                if host.startswith("127")
                else "0.0.0.0"  # noqa: S104 Possible binding to all interfaces
                if host == "0.0.0.0"  # noqa: S104 Possible binding to all interfaces
                else f"{host.split('.')[0]}.0.0.0",
                "port": port,
                "workers": workers,
                **prepare_profile_for_telemetry(profile),
                "reload": reload,
                "tunnel": tunnel,
            }
        )
        await listen(
            host=host,
            port=port,
            workers=workers,
            profile=profile,
            tunnel=tunnel,
            ngrok_token=ngrok_token,
            ngrok_domain=ngrok_domain,
            reload=reload,
        )

    anyio.run(_serve)



import anyio
from rich import print

from unpage.agent.utils import get_agent_templates
from unpage.cli.agent._app import agent_app
from unpage.telemetry import client as telemetry


@agent_app.command()
def templates() -> None:
    """List the available agent templates."""

    async def _run() -> None:
        await telemetry.send_event(
            {
                "command": "agent templates",
            }
        )
        print("Available agent templates:")
        for template in sorted(get_agent_templates()):
            print(f"* {template}")

    anyio.run(_run)



import anyio

from unpage.cli.graph._app import graph_app
from unpage.cli.graph._background import (
    cleanup_pid_file,
    get_log_file,
    get_pid_file,
    is_process_running,
)
from unpage.cli.options import PROFILE_OPTION
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@graph_app.command()
def status(profile: str = PROFILE_OPTION) -> None:
    """Check if graph build is running"""

    async def _status() -> None:
        await telemetry.send_event(
            {
                "command": "graph status",
                **prepare_profile_for_telemetry(profile),
            }
        )
        _check_status()

    def _check_status() -> None:
        pid_file = get_pid_file(profile)

        if not pid_file.exists():
            print(f"No graph build running for profile '{profile}'")
            return

        try:
            pid = int(pid_file.read_text().strip())
            if is_process_running(pid):
                print(f"Graph build running for profile '{profile}' (PID: {pid})")

                # Show log file info if it exists
                log_file = get_log_file(profile)
                if log_file.exists():
                    print(f"View logs: unpage graph logs --profile {profile} --follow")
            else:
                print(f"Stale PID file found for profile '{profile}', cleaning up...")
                cleanup_pid_file(profile)
        except ValueError:
            print(f"Corrupted PID file found for profile '{profile}', cleaning up...")
            cleanup_pid_file(profile)

    anyio.run(_status)



# SPDX-License-Identifier: MIT

from typing import Optional, Union

import redis

from drain3.persistence_handler import PersistenceHandler


class RedisPersistence(PersistenceHandler):
    def __init__(self,
                 redis_host: str,
                 redis_port: int,
                 redis_db: int,
                 redis_pass: Optional[str],
                 is_ssl: bool,
                 redis_key: Union[bytes, str, memoryview]) -> None:
        self.redis_host = redis_host
        self.redis_port = redis_port
        self.redis_db = redis_db
        self.redis_pass = redis_pass
        self.is_ssl = is_ssl
        self.redis_key = redis_key
        self.r = redis.Redis(host=self.redis_host,
                             port=self.redis_port,
                             db=self.redis_db,
                             password=self.redis_pass,
                             ssl=self.is_ssl)

    def save_state(self, state: bytes) -> None:
        self.r.set(self.redis_key, state)

    def load_state(self) -> Optional[bytes]:
        return self.r.get(self.redis_key)



# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from typing import Optional


class PersistenceHandler(ABC):

    @abstractmethod
    def save_state(self, state: bytes) -> None:
        pass

    @abstractmethod
    def load_state(self) -> Optional[bytes]:
        pass



import anyio
import rich

from unpage.cli.mcp.client.logs.app import client_logs_app
from unpage.telemetry import client as telemetry


@client_logs_app.command()
def claude() -> None:
    """Show logs for Claude Desktop Unpage MCP Server"""

    async def _run() -> None:
        rich.print("Logs are available at:")
        rich.print("~/Library/Logs/Claude/mcp.log")
        rich.print("~/Library/Logs/Claude/mcp-server-unpage.log")
        await telemetry.send_event(
            {
                "command": "mcp client logs claude",
            }
        )

    anyio.run(_run)



# SPDX-License-Identifier: MIT

from typing import Optional

from drain3.persistence_handler import PersistenceHandler


class MemoryBufferPersistence(PersistenceHandler):
    def __init__(self) -> None:
        self.state: Optional[bytes] = None

    def save_state(self, state: bytes) -> None:
        self.state = state

    def load_state(self) -> Optional[bytes]:
        return self.state


import typer

from unpage.cli.agent._app import agent_app
from unpage.cli.graph._app import graph_app
from unpage.cli.mcp._app import mcp_app
from unpage.cli.mlflow._app import mlflow_app
from unpage.warnings import filter_all_warnings

filter_all_warnings()

app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]}, no_args_is_help=True)
app.add_typer(mcp_app, name="mcp")
app.add_typer(graph_app, name="graph")
app.add_typer(agent_app, name="agent")
app.add_typer(mlflow_app, name="mlflow")



import sys
from typing import TYPE_CHECKING, cast

import anyio
import typer
from rich import print

from unpage.agent.analysis import AnalysisAgent
from unpage.agent.utils import load_agent
from unpage.cli.agent._app import agent_app
from unpage.cli.options import PROFILE_OPTION
from unpage.config.utils import load_config
from unpage.plugins.base import PluginManager
from unpage.telemetry import client as telemetry
from unpage.telemetry import hash_value, prepare_profile_for_telemetry

if TYPE_CHECKING:
    from unpage.plugins.pagerduty.plugin import PagerDutyPlugin


@agent_app.command(
    help="Run an agent with the provided payload and print the analysis. A payload can be passed as an argument or piped to stdin."
)
def run(
    agent_name: str = typer.Argument(..., help="The name of the agent to run"),
    payload: str | None = typer.Argument(
        None,
        help="The alert payload to analyze. Alternatively, you can pipe the payload to stdin.",
    ),
    pagerduty_incident: str | None = typer.Option(
        None, help="PagerDuty incident ID or URL to use instead of payload or stdin"
    ),
    profile: str = PROFILE_OPTION,
    debug: bool = typer.Option(
        False,
        help="Enable debug mode to print the history of the agent.",
    ),
) -> None:
    """Run an agent with the provided payload and print the analysis."""

    async def _run() -> None:
        await telemetry.send_event(
            {
                "command": "agent run",
                **prepare_profile_for_telemetry(profile),
                "agent_name_sha256": hash_value(agent_name),
                "debug": debug,
                "has_payload": payload is not None,
                "has_pagerduty_incident": bool(pagerduty_incident),
            }
        )
        plugin_manager = PluginManager(load_config(profile, create=False))
        data = ""
        if pagerduty_incident:
            if payload is not None or not sys.stdin.isatty():
                print("[red]Cannot pass --pagerduty-incident with --payload or stdin.[/red]")
                raise typer.Exit(code=1)
            incident_id = pagerduty_incident
            if "/" in pagerduty_incident:
                incident_id = [x for x in pagerduty_incident.split("/") if x][-1]
            pd = cast("PagerDutyPlugin", plugin_manager.get_plugin("pagerduty"))
            incident = await pd.get_incident_by_id(incident_id)
            data = incident.model_dump_json()

        # Read data from stdin if it's being piped to us.
        if not data and not sys.stdin.isatty():
            if payload is not None:
                print("[red]Cannot pass a payload argument when piping data to stdin.[/red]")
                raise typer.Exit(code=1)
            data = sys.stdin.read().strip()
        elif not data:
            # Otherwise, use the payload argument.
            data = payload

        if not data:
            print("[red]No payload provided.[/red]")
            print(
                "[bold]Pass an alert payload as an argument or pipe the payload data to stdin.[/bold]"
            )
            raise typer.Exit(code=1)

        # Get the config directory and load the specific agent
        try:
            agent = load_agent(agent_name, profile)
        except FileNotFoundError as ex:
            print(f"[red]Agent {agent_name!r} not found at {str(ex.filename)!r}[/red]")
            print(f"[bold]Use 'unpage agent create {agent_name!r}' to create a new agent.[/bold]")
            raise typer.Exit(code=1) from ex

        # Run the analysis with the specific agent
        analysis_agent = AnalysisAgent(profile)
        try:
            result = await analysis_agent.acall(payload=data, agent=agent)
            print(result)
        except Exception as ex:
            print(f"[red]Analysis failed:[/red] {ex}")
            raise typer.Exit(code=1) from ex
        finally:
            if debug:
                print("\n\n===== DEBUG OUTPUT =====\n")
                analysis_agent.inspect_history(n=1000)
                print("\n===== END DEBUG OUTPUT =====\n\n")

    anyio.run(_run)



import asyncio
import shutil

import anyio
import typer

from unpage.cli.graph._app import graph_app
from unpage.cli.graph._background import get_log_file
from unpage.cli.options import PROFILE_OPTION
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@graph_app.command()
def logs(
    profile: str = PROFILE_OPTION,
    follow: bool = typer.Option(False, "--follow", "-f", help="Follow log output"),
) -> None:
    """View graph build logs"""

    async def _run() -> None:
        await telemetry.send_event(
            {
                "command": "graph logs",
                **prepare_profile_for_telemetry(profile),
                "follow": follow,
            }
        )
        log_file = get_log_file(profile)
        tail_cmd = shutil.which("tail")
        if not tail_cmd:
            print("'tail' command not found. Please install it.")
            return

        if not log_file.exists():
            print(f"No log file found for profile '{profile}'")
            print(f"Expected location: {log_file}")
            return

        if follow:
            print(f"Following logs for profile '{profile}' (Ctrl+C to stop)")
            print(f"Log file: {log_file}")

            try:
                proc = await asyncio.create_subprocess_shell(f"{tail_cmd} -f {log_file!s}")
                await proc.wait()
            except KeyboardInterrupt:
                print("\nStopped following logs")
        else:
            print(f"Recent logs for profile '{profile}':")
            print(f"Log file: {log_file}")

            proc = await asyncio.create_subprocess_shell(f"{tail_cmd} -n 50 {log_file!s}")
            await proc.wait()

    anyio.run(_run)



import typer

from unpage.cli.mcp.client._app import client_app
from unpage.cli.mcp.tools._app import tools_app

mcp_app = typer.Typer(help="MCP tool commands", no_args_is_help=True)
mcp_app.add_typer(tools_app, name="tools")
mcp_app.add_typer(client_app, name="client")



import anyio
from rich import print

from unpage.agent.utils import get_agents
from unpage.cli.agent._app import agent_app
from unpage.cli.options import PROFILE_OPTION
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@agent_app.command()
def list(profile: str = PROFILE_OPTION) -> None:
    """List the available agents."""

    async def _run() -> None:
        await telemetry.send_event(
            {
                "command": "agent list",
                **prepare_profile_for_telemetry(profile),
            }
        )
        print("Available agents:")
        for agent in sorted(get_agents(profile)):
            print(f"* {agent}")

    anyio.run(_run)



import os
from typing import Any

import litellm
import questionary
from litellm import acompletion

from unpage.config.utils import PluginSettings
from unpage.plugins.base import Plugin
from unpage.utils import classproperty, select


class LlmPlugin(Plugin):
    """A plugin for configuring LLM models."""

    def __init__(
        self,
        *args: Any,
        model: str,
        api_key: str,
        temperature: float,
        max_tokens: int,
        cache: bool,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.model = model
        self.api_key = api_key
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.cache = cache
        if (
            self.model.startswith("bedrock/")
            and not os.environ.get("AWS_REGION")
            and not os.environ.get("AWS_DEFAULT_REGION")
        ):
            os.environ["AWS_REGION"] = "us-east-1"

    @classproperty
    def default_plugin_settings(cls) -> PluginSettings:
        return {
            "model": "openai/gpt-4o-mini",
            "api_key": "",
            "temperature": 0,
            "max_tokens": 8192,
            "cache": True,
        }

    async def interactive_configure(self) -> PluginSettings:
        """Interactive wizard for configuring the settings of this plugin."""
        defaults = self.default_plugin_settings
        model_list = [
            model if model.startswith(f"{provider}/") else f"{provider}/{model}"
            for provider, models in litellm.models_by_provider.items()
            for model in models
        ]
        return {
            "model": await select(
                "Model",
                choices=model_list,
                default=self.model or defaults["model"],
                use_jk_keys=False,
                use_search_filter=True,
            ),
            "api_key": await questionary.password(
                "API key",
                default=self.api_key or defaults["api_key"],
            ).unsafe_ask_async(),
            "temperature": self.temperature or defaults["temperature"],
            "max_tokens": self.max_tokens or defaults["max_tokens"],
            "cache": self.cache or defaults["cache"],
        }

    async def validate_plugin_config(self) -> None:
        params = {
            "model": self.model,
            "api_key": self.api_key,
            **({"temperature": self.temperature} if not self.model.startswith("bedrock/") else {}),
            "max_tokens": self.max_tokens,
            "cache": self.cache,
        }
        await acompletion(
            **params,
            messages=[
                {
                    "role": "user",
                    "content": "hiiiii",
                }
            ],
        )



import anyio

from unpage.agent.utils import delete_agent
from unpage.cli.agent._app import agent_app
from unpage.cli.options import PROFILE_OPTION
from unpage.telemetry import client as telemetry
from unpage.telemetry import hash_value, prepare_profile_for_telemetry


@agent_app.command()
def delete(agent_name: str, profile: str = PROFILE_OPTION) -> None:
    """Delete an agent."""

    async def _run() -> None:
        await telemetry.send_event(
            {
                "command": "agent delete",
                "agent_name_sha256": hash_value(agent_name),
                **prepare_profile_for_telemetry(profile),
            }
        )
        delete_agent(agent_name, profile)

    anyio.run(_run)



# SPDX-License-Identifier: Apache-2.0
# Based on https://github.com/davidohana/SimpleProfiler/blob/main/python/simple_profiler.py

import os
import time

from abc import ABC, abstractmethod
from typing import Any, Callable, MutableMapping, Union


class Profiler(ABC):

    @abstractmethod
    def start_section(self, section_name: str) -> None:
        pass

    @abstractmethod
    def end_section(self, section_name: str = "") -> None:
        pass

    @abstractmethod
    def report(self, period_sec: int = 30) -> None:
        pass


class NullProfiler(Profiler):
    """A no-op profiler. Use it instead of SimpleProfiler in case you want to disable profiling."""

    def start_section(self, section_name: str) -> None:
        pass

    def end_section(self, section_name: str = "") -> None:
        pass

    def report(self, period_sec: int = 30) -> None:
        pass


class SimpleProfiler(Profiler):
    def __init__(self,
                 reset_after_sample_count: int = 0,
                 enclosing_section_name: str = "total",
                 printer: Callable[[str], Any] = print,
                 report_sec: int = 30):
        self.printer = printer
        self.enclosing_section_name = enclosing_section_name
        self.reset_after_sample_count = reset_after_sample_count
        self.report_sec = report_sec

        self.section_to_stats: MutableMapping[str, ProfiledSectionStats] = {}
        self.last_report_timestamp_sec = time.time()
        self.last_started_section_name = ""

    def start_section(self, section_name: str) -> None:
        """Start measuring a section"""

        if not section_name:
            raise ValueError("Section name is empty")
        self.last_started_section_name = section_name

        section = self.section_to_stats.get(section_name, None)
        if section is None:
            section = ProfiledSectionStats(section_name)
            self.section_to_stats[section_name] = section

        if section.start_time_sec != 0:
            raise ValueError(f"Section {section_name} is already started")

        section.start_time_sec = time.time()

    def end_section(self, name: str = "") -> None:
        """End measuring a section. Leave section name empty to end the last started section."""

        now = time.time()

        section_name = name
        if not name:
            section_name = self.last_started_section_name

        if not section_name:
            raise ValueError("Neither section name is specified nor a section is started")

        if section_name not in self.section_to_stats:
            raise ValueError(f"Section {section_name} does not exist")
        section = self.section_to_stats[section_name]

        if section.start_time_sec == 0:
            raise ValueError(f"Section {section_name} was not started")

        took_sec = now - section.start_time_sec
        if 0 < self.reset_after_sample_count == section.sample_count:
            section.sample_count_batch = 0
            section.total_time_sec_batch = 0

        section.sample_count += 1
        section.total_time_sec += took_sec
        section.sample_count_batch += 1
        section.total_time_sec_batch += took_sec
        section.start_time_sec = 0

    def report(self, period_sec: int = 30) -> None:
        """Print results using [printer] function. By default prints to stdout."""
        if time.time() - self.last_report_timestamp_sec < period_sec:
            return

        enclosing_time_sec: Union[int, float] = 0
        if self.enclosing_section_name:
            if self.enclosing_section_name in self.section_to_stats:
                enclosing_time_sec = self.section_to_stats[self.enclosing_section_name].total_time_sec

        include_batch_rates = self.reset_after_sample_count > 0

        sections = self.section_to_stats.values()
        sorted_sections = sorted(sections, key=lambda it: it.total_time_sec, reverse=True)
        lines = map(lambda it: it.to_string(enclosing_time_sec, include_batch_rates), sorted_sections)
        text = os.linesep.join(lines)
        self.printer(text)

        self.last_report_timestamp_sec = time.time()


class ProfiledSectionStats:
    def __init__(self, section_name: str, start_time_sec: Union[int, float] = 0, sample_count: int = 0,
                 total_time_sec: Union[int, float] = 0, sample_count_batch: int = 0,
                 total_time_sec_batch: Union[int, float] = 0) -> None:
        self.section_name = section_name
        self.start_time_sec = start_time_sec
        self.sample_count = sample_count
        self.total_time_sec = total_time_sec
        self.sample_count_batch = sample_count_batch
        self.total_time_sec_batch = total_time_sec_batch

    def to_string(self, enclosing_time_sec: Union[int, float], include_batch_rates: bool) -> str:
        took_sec_text = f"{self.total_time_sec:>8.2f} s"
        if enclosing_time_sec > 0:
            took_sec_text += f" ({100 * self.total_time_sec / enclosing_time_sec:>6.2f}%)"

        ms_per_k_samples = f"{1000000 * self.total_time_sec / self.sample_count: 7.2f}"

        if self.total_time_sec > 0:
            samples_per_sec = f"{self.sample_count / self.total_time_sec: 15,.2f}"
        else:
            samples_per_sec = "N/A"

        if include_batch_rates:
            ms_per_k_samples += f" ({1000000 * self.total_time_sec_batch / self.sample_count_batch: 7.2f})"
            if self.total_time_sec_batch > 0:
                samples_per_sec += f" ({self.sample_count_batch / self.total_time_sec_batch: 15,.2f})"
            else:
                samples_per_sec += " (N/A)"

        return f"{self.section_name: <15}: took {took_sec_text}, " \
               f"{self.sample_count: >10,} samples, " \
               f"{ms_per_k_samples} ms / 1000 samples, " \
               f"{samples_per_sec} hz"
